/*
 * getut.c -- read utmp file
 *
 * Copyright (C) 1986, 1996  Philip L. Budne
 *
 * This file is part of "Phil's Finger Program".
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 */

# ifndef lint
static char *rcsid = "$Id: getut.c,v 3.7 1999/03/02 04:46:08 phil Exp $";
# endif /* lint not defined */

# include "finger.h"

# include <stdio.h>

# include <sys/types.h>			/* for time_t (in utmp) */

# ifdef HAVE_UTMPX
# include <utmpx.h>
# define utmp utmpx
# define getutent getutxent
# define endutent endutxent
# define UTFILE UTMPX_FILE
# define ut_time ut_tv.tv_sec
# else  /* HAVE_UTMPX not defined */
# include <utmp.h>
# ifdef UTMP_FILE
# define UTFILE UTMP_FILE
# else  /* UTMP_FILE not defined */
# ifdef _PATH_UTMP
# define UTFILE _PATH_UTMP
# else  /* _PATH_UTMP not defined */
# define UTFILE "/etc/utmp"
# endif /* _PATH_UTMP not defined */
# endif /* UTMP_FILE not defined */
# endif /* HAVE_UTMPX not defined */

# ifdef USER_PROCESS
# include <errno.h>
# endif /* USER_PROCESS defined */

# ifdef UTMP_BOOTTIME
extern time_t bootsec;
# endif /* UTMP_BOOTTIME defined */

# define MAXNAME 100

LOCAL char utfile[MAXNAME] = UTFILE;
LOCAL FILE *f = NULL;

void setutent() {
    if( f == NULL )
	f = fopen( utfile, "r" );
    else
	rewind( f );
} /* setutent */

void endutent() {
    if( f != NULL )
	fclose( f );
} /* endutent */

# if 0
void utmpname( file )
char *file;
{
    endutent();
    strcpy(utfile, file);
} /* utmpname */
# endif /* 0 */

struct utmp *getutent() {
    static struct utmp ut;
 
    if( f == NULL ) {
	setutent();
	if( f == NULL )
	    return( NULL );
    }

    while( fread(&ut, sizeof( ut ), 1, f) == 1 ) {
# ifdef USER_PROCESS
	if( ut.ut_type == USER_PROCESS
	   && (kill( ut.ut_pid, 0 ) == 0 || errno == EPERM ) )
# endif /* USER_PROCESS defined */
	    return( &ut );
# ifdef UTMP_BOOTTIME
	if( ut.ut_type == BOOT_TIME)
	    bootsec = ut.ut_time;
# endif /* UTMP_BOOTTIME defined */
    }
    rewind( f );
    return( NULL );
} /* getutent */

/*
 *	not implemented:
 *
 *	void pututline( struct utmp *u );
 *		(uses direct write(2))
 *	struct utmp *getutline( struct utmp *line )
 *		find entry with matching ut->line
 *	struct utmp *getutid( struct utmp *id )
 *		find entry with matching ut->type (!!)
 *		(RUN_LVL, BOOT_TIME, OLD_TIME, NEW_TIME)
 *		if one of (INIT_PROCESS, LOGIN_PROCESS, USER_PROCESS
 *		or DEAD_PROCESS) ut_id must mach also.
 */

/*
 * Local variables:
 * comment-column: 40
 * End:
 */
