/*
 * lastlog.c - get time and location of last login
 *
 * Copyright (C) 1986, 1996  Philip L. Budne
 *
 * This file is part of "Phil's Finger Program".
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 */

# ifndef lint
static char *rcsid = "$Id: lastlog.c,v 3.5 2000/02/18 06:04:07 phil Exp $";
# endif /* lint not defined */

# include "finger.h"
# ifdef LASTLOG
# include <sys/types.h>
# include <ctype.h>

# ifdef LASTLOG_UTMP
# include <utmp.h>
# define LLPATH _PATH_LASTLOG
# else  /* LASTLOG_UTMP not defined */
# include <lastlog.h>

# ifndef LLPATH
# define LLPATH "/usr/adm/lastlog"
# endif /* LLPATH not defined */
# endif /* LASTLOG_UTMP not defined */

# ifndef LASTLOG_DIR
LOCAL llfile = -1;
# endif /* LASTLOG_DIR not defined */

GLOBAL void llopen() {
# ifndef LASTLOG_DIR
    if( llfile < 0 )
	llfile = open(LLPATH, 0);
    else
	lseek( llfile, 0L, 0 );
# endif /* LASTLOG_DIR not defined */
} /* llopen */

GLOBAL void llclose() {
# ifndef LASTLOG_DIR
    if( llfile >= 0 )
	close( llfile );
# endif /* LASTLOG_DIR not defined */
} /* llclose */

GLOBAL BOOL lltime( uid, uname, location, lsize, time, host )
int uid;
char *uname;
char *location;
int lsize;				/* size of location buffer */
time_t *time;
int *host;
{
    struct lastlog ll;
# ifdef LASTLOG_DIR
    char fname[128];
    int llfile;

    sprintf(fname, "%s/%s", LLPATH, uname );
    llfile = open( fname, 0);
    if( llfile < 0 )
	return( FALSE );
# else  /* LASTLOG_DIR not defined */
    if( llfile < 0 )
	return( FALSE );

    lseek( llfile, uid * sizeof( ll ), 0 );
# endif /* LASTLOG_DIR not defined */

    if( read(llfile, &ll, sizeof( ll ) ) != sizeof( ll ) ) {
# ifdef LASTLOG_DIR
	close(llfile);
# endif /* LASTLOG_DIR defined */
	return( FALSE );
    }

    *time = ll.ll_time;			/* return time of login */
    if( ll.ll_host[0] != EOS ) {
	/* IRIX has remote username in ll_line?? */
	*host = TRUE;
	strlzcpy(location, sizeof(location),
		ll.ll_host, sizeof( ll.ll_host ) ); /* copy fixed len */

	if( isdigit(*location) )
	    checkhost( location, lsize, FALSE ); /* hack octets */
	undomain( location, 0 ); 	/* remove domains, but not prefixes */
	HZUP( location );		/* upper caseify */
    }
    else {
	*host = FALSE;
	strlzcpy(location, sizeof(location),
		 ll.ll_line, sizeof( ll.ll_line ) );
    }
# ifdef LASTLOG_DIR
    close( llfile );
# endif /* LASTLOG_DIR defined */
    return( TRUE );
} /* lltime */
# endif /* LASTLOG defined */
/*
 * Local variables:
 * comment-column: 40
 * End:
 */
